import glob
import logging

import click
import prettytable
from astropy import coordinates as coord
from astropy import time
from astropy.io import fits

logger = logging.getLogger(__name__)


@click.command(
    epilog="""Check out the documentation at
                https://pyscope.readthedocs.io/ for more
                information."""
)
@click.option("-d", "--date", default="", help="Date [default all].")
@click.option("-f", "--filt", default="", help="Filter name [default all].")
@click.option("-r", "--readout", default="", help="Readout mode [default all].")
@click.option("-b", "--binning", default="", help="Binning [default all].")
@click.option(
    "-e",
    "--exptime",
    default="",
    help=f"""Approximate exposure time [default all].
                Note that an error of up to 1% is permitted to allow for imprecisions
                in the camera.""",
)
@click.option("-t", "--target", default="", help="Target name [default all].")
@click.option(
    "-v", "--verbose", count=True, type=click.IntRange(0, 1), help="Verbose output."
)
@click.argument("fnames", nargs=-1, type=click.Path(exists=True, dir_okay=False))
@click.version_option()
def fitslist_cli(
    check_date,
    check_filt,
    check_readout,
    check_binning,
    check_exptime,
    check_target,
    verbose,
    fnames,
):
    """List FITS files and their properties."""

    # Set up logging
    logger.setLevel(int(10 * (1 - verbose)))
    logger.debug(
        f"""filt={check_filt}, readout={check_readout}, binning={check_binning}, exptime={check_exptime},
                target={check_target}, verbose={verbose}, fnames={fnames}"""
    )

    # Get list of files
    ftsfiles = []
    if len(fnames) == 0:
        logger.debug("No arguments passed. Using glob to find files.")
        for ext in (".fts", ".fits", ".fit"):
            ftsfiles.extend(glob.glob(f"./*{ext}"))
    else:
        ftsfiles = fnames
    logger.debug(f"fnames={fnames}")
    logger.debug(f"Found {len(fnames)} files.")

    print_rows = []
    for ftsfile in ftsfiles:
        try:
            with fits.open(ftsfile) as hdu:
                header = hdu[0].header
                data = hdu[0].data
        except:
            logger.warning(f"Could not open {ftsfile}.")
            continue

        # Get properties
        date = time.Time(header["DATE-OBS"], format="fits", scale="utc")
        if date.strftime("%Y-%m-%d") not in check_date.split(",") or check_date == "":
            logger.debug(
                f"Date {date.strftime('%Y-%m-%d')} not in {check_date}. Skipping {ftsfile}."
            )
            continue

        # Filter
        try:
            filt = header["FILTER"]
        except KeyError:
            try:
                filt = header["FILT"]
            except KeyError:
                filt = ""
        if filt not in check_filt.split(",") or check_filt == "":
            logger.debug(f"Filter {filt} not in {check_filt}. Skipping {ftsfile}.")
            continue

        # Readout mode
        try:
            readout_mode = header["READOUTM"]
        except KeyError:
            try:
                readout_mode = header["READOUT"]
            except KeyError:
                readout_mode = ""
        if readout_mode not in check_readout.split(",") or check_readout == "":
            logger.debug(
                f"Readout mode {readout_mode} not in {check_readout}. Skipping {ftsfile}."
            )
            continue

        # Binning
        try:
            x_binning = header["XBINNING"]
            y_binning = header["YBINNING"]
        except KeyError:
            x_binning = ""
            y_binning = ""
        if (
            x_binning + "x" + y_binning not in check_binning.split(",")
            or check_binning == ""
        ):
            logger.debug(
                f"Binning {x_binning}x{y_binning} not in {check_binning}. Skipping {ftsfile}."
            )
            continue

        # Exposure time
        try:
            exptime = header["EXPTIME"]
        except KeyError:
            try:
                exptime = header["EXPOSURE"]
            except KeyError:
                exptime = -1
        if check_exptime != "":
            for c_exp in check_exptime.split(","):
                if exptime < float(c_exp) * 0.99 or exptime > float(c_exp) * 1.01:
                    logger.debug(
                        f"Exposure time {exptime} not in {check_exptime}. Skipping {ftsfile}."
                    )
                    continue

        # Target
        try:
            target_name = header["OBJECT"]
        except KeyError:
            try:
                target_name = header["OBJNAME"]
            except KeyError:
                try:
                    target_name = header["SOURCE"]
                except KeyError:
                    try:
                        target_name = header["TARGNAME"]
                    except KeyError:
                        target_name = ""
        if target_name not in check_target.split(",") or check_target == "":
            logger.debug(
                f"Target {target_name} not in {check_target}. Skipping {ftsfile}."
            )
            continue

        # Actual coordinates
        try:
            ra = header["OBJCTRA"]
            dec = header["OBJCTDEC"]
        except KeyError:
            try:
                ra = header["OBJRA"]
                dec = header["OBJDEC"]
            except KeyError:
                try:
                    logger.warning("No coordinates found. Using telescope coordinates.")
                    ra = header["TELRA"]
                    dec = header["TELDEC"]
                except KeyError:
                    ra = ""
                    dec = ""
        if "" not in (ra, dec):
            obj = coord.SkyCoord(ra, dec)

        # Scheduled coordinates
        try:
            sched_ra = header["SCHEDRA"]
            sched_dec = header["SCHEDDEC"]
        except KeyError:
            sched_ra = ""
            sched_dec = ""
        if "" not in (sched_ra, sched_dec):
            sched_obj = coord.SkyCoord(sched_ra, sched_dec)

        if "" not in (ra, dec, sched_ra, sched_dec):
            dra = (obj.ra - sched_obj.ra).to("arcsec")
            ddec = (obj.dec - sched_obj.dec).to("arcsec")
        else:
            dra = ""
            ddec = ""

        # ZP
        try:
            zp = header["ZMAG"]
            zp_err = header["ZMAGERR"]
        except KeyError:
            zp = ""
            zp_err = ""

        # FWHM
        try:
            fwhmh = header["FWHMH"]
            fwhmhs = header["FWHMHS"]
            fwhmv = header["FWHMV"]
            fwhmvs = header["FWHMVS"]
        except KeyError:
            fwhmh = ""
            fwhmhs = ""
            fwhmv = ""
            fwhmvs = ""

        # Moon
        try:
            moon_angle = header["MOONANGL"]
            moon_phs = header["MOONPHAS"]
        except KeyError:
            moon_angle = ""
            moon_phs = ""

        print_rows.append(
            [
                fitsfile,
                target_name,
                date.jd,
                date.iso,
                filt,
                readout_mode,
                x_binning + "x" + y_binning,
                f"{exptime:.3f}",
                obj.ra.hms,
                obj.dec.dms,
                f"{zp:.3f}+/-{zp_err:.3f}",
                f"{fwhmh:.2f}+/-{fwhmhs:.2f}",
                f"{fwhmv:.2f}+/-{fwhmvs:.2f}",
                f"{moon_angle:.1f}",
                f"{moon_phs:.1f}",
                f"{dra:.2f}",
                f"{ddec:.2f}",
            ]
        )

    logger.debug(f"Found {len(print_rows)} files matching criteria.")

    logger.debug("Sorting by JD...")
    print_rows.sort(key=lambda x: x[2])

    table = prettytable.PrettyTable()
    table.add_rows(print_rows)
    table.set_style(prettytable.SINGLE_BORDER)
    table.field_names = [
        "FITS file",
        "Target",
        "JD",
        "UT",
        "Filter",
        "Readout",
        "Binning",
        "Exp. time [s]",
        "RA",
        "Dec",
        "ZP",
        "FWHM H [pix]",
        "FWHM V [pix]",
        "Moon angle [deg]",
        "Moon phase [0-1]",
        "dRA [arcsec]",
        "dDec [arcsec]",
    ]

    T.align["FITS file"] = "l"
    click.echo(table)
    click.echo()
    click.echo(f"Number of images = {len(print_rows)}")

    fwhmh = np.median([row[11] for row in print_rows])
    fwhmhs = np.std([row[11] for row in print_rows])
    fwhmv = np.median([row[12] for row in print_rows])
    fwhmvs = np.std([row[12] for row in print_rows])
    click.echo(f"Median FWHM H = {fwhmh:.2f} +/- {fwhmhs:.2f} pix")
    click.echo(f"Median FWHM V = {fwhmv:.2f} +/- {fwhmvs:.2f} pix")

    moon_phs = np.mean([row[14] for row in print_rows])
    click.echo(f"Mean Moon phase = {moon_phs:.2f}")

    click.echo("\nMedian zero-point magnitudes")
    table = prettytable.PrettyTable()
    table.set_style(prettytable.SINGLE_BORDER)
    table.field_names = ["Readout", "Filter", "ZP"]
    for filt in ["u", "b", "g", "r", "i", "z"]:
        rows = [row for row in print_rows if row[4] == filt]
        if len(rows) == 0:
            continue
        for readout in np.unique([row[5] for row in rows]):
            zp = np.median([row[10] for row in rows if row[5] == readout])
            zp_err = np.std([row[10] for row in rows if row[5] == readout])
            table.add_row([readout, filt, f"{zp:.3f}+/-{zp_err:.3f}"])
    click.echo(table)

    return print_rows


fitslist = fitslist_cli.callback



import pytest


def test_focusoffsets(device, disconnect):
    assert device.FocusOffsets is not None


def test_names(device, disconnect):
    assert device.Names is not None


def test_position(device, disconnect):
    assert device.Position is not None
    device.Position = 0
    assert device.Position == 0



from abc import ABCMeta


class _DocstringInheritee(ABCMeta):
    def __new__(mcls, classname, bases, cls_dict):
        cls = super().__new__(mcls, classname, bases, cls_dict)
        for name, member in cls_dict.items():
            if not getattr(member, "__doc__"):
                try:
                    member.__doc__ = getattr(bases[-1], name).__doc__
                except AttributeError:
                    pass
        return cls



import platform

import pytest

from pyscope.observatory import IPCoverCalibrator


@pytest.mark.skipif(platform.node() != "TCC1-MACRO", reason="Only run on TCC1-MACRO")
def test_IPCoverCalibrator():
    c = IPCoverCalibrator("192.168.2.22", 2101, 1024)

    c.CalibratorOn(1)
    c.CalibratorOff()

    with pytest.raises(NotImplementedError):
        c.CloseCover()
    with pytest.raises(NotImplementedError):
        c.HaltCover()
    with pytest.raises(NotImplementedError):
        c.OpenCover()

    assert c.Brightness is None
    assert c.CalibratorState is None
    assert c.CoverState is None
    assert c.MaxBrightness == 254

    assert c.tcp_ip is not None
    assert c.tcp_port is not None
    assert c.buffer_size is not None



from abc import ABC, abstractmethod

from ._docstring_inheritee import _DocstringInheritee


class Telescope(ABC, metaclass=_DocstringInheritee):
    @abstractmethod
    def __init__(self, *args, **kwargs):
        pass

    @abstractmethod
    def AbortSlew(self):
        pass

    @abstractmethod
    def AxisRates(self, Axis):
        pass

    @abstractmethod
    def CanMoveAxis(self, Axis):
        pass

    @abstractmethod
    def DestinationSideOfPier(self, RightAscension, Declination):
        pass

    @abstractmethod
    def FindHome(self):
        pass

    @abstractmethod
    def MoveAxis(self, Axis, Rate):
        pass

    @abstractmethod
    def Park(self):
        pass

    @abstractmethod
    def PulseGuide(self, Direction, Duration):
        pass

    @abstractmethod
    def SetPark(self):
        pass

    @abstractmethod
    def SlewToAltAz(self, Azimuth, Altitude):
        pass

    @abstractmethod
    def SlewToAltAzAsync(self, Azimuth, Altitude):
        pass

    @abstractmethod
    def SlewToCoordinates(self, RightAscension, Declination):
        pass

    @abstractmethod
    def SlewToCoordinatesAsync(self, RightAscension, Declination):
        pass

    @abstractmethod
    def SlewToTarget(self):
        pass

    @abstractmethod
    def SlewToTargetAsync(self):
        pass

    @abstractmethod
    def SyncToAltAz(self, Azimuth, Altitude):
        pass

    @abstractmethod
    def SyncToCoordinates(self, RightAscension, Declination):
        pass

    @abstractmethod
    def SyncToTarget(self):
        pass

    @abstractmethod
    def Unpark(self):
        pass

    @property
    @abstractmethod
    def AlignmentMode(self):
        pass

    @AlignmentMode.setter
    @abstractmethod
    def AlignmentMode(self, value):
        pass

    @property
    @abstractmethod
    def Altitude(self):
        pass

    @property
    @abstractmethod
    def ApertureArea(self):
        pass

    @property
    @abstractmethod
    def ApertureDiameter(self):
        pass

    @property
    @abstractmethod
    def AtHome(self):
        pass

    @property
    @abstractmethod
    def AtPark(self):
        pass

    @property
    @abstractmethod
    def Azimuth(self):
        pass

    @property
    @abstractmethod
    def CanFindHome(self):
        pass

    @property
    @abstractmethod
    def CanPark(self):
        pass

    @property
    @abstractmethod
    def CanPulseGuide(self):
        pass

    @property
    @abstractmethod
    def CanSetDeclinationRate(self):
        pass

    @property
    @abstractmethod
    def CanSetGuideRates(self):
        pass

    @property
    @abstractmethod
    def CanSetPark(self):
        pass

    @property
    @abstractmethod
    def CanSetPierSide(self):
        pass

    @property
    @abstractmethod
    def CanSetRightAscensionRate(self):
        pass

    @property
    @abstractmethod
    def CanSetTracking(self):
        pass

    @property
    @abstractmethod
    def CanSlew(self):
        pass

    @property
    @abstractmethod
    def CanSlewAltAz(self):
        pass

    @property
    @abstractmethod
    def CanSlewAltAzAsync(self):
        pass

    @property
    @abstractmethod
    def CanSlewAsync(self):
        pass

    @property
    @abstractmethod
    def CanSync(self):
        pass

    @property
    @abstractmethod
    def CanSyncAltAz(self):
        pass

    @property
    @abstractmethod
    def CanUnpark(self):
        pass

    @property
    @abstractmethod
    def Declination(self):
        pass

    @Declination.setter
    @abstractmethod
    def Declination(self, value):
        pass

    @property
    @abstractmethod
    def DeclinationRate(self):
        pass

    @DeclinationRate.setter
    @abstractmethod
    def DeclinationRate(self, value):
        pass

    @property
    @abstractmethod
    def DoesRefraction(self):
        pass

    @DoesRefraction.setter
    @abstractmethod
    def DoesRefraction(self, value):
        pass

    @property
    @abstractmethod
    def EquatorialSystem(self):
        pass

    @property
    @abstractmethod
    def FocalLength(self):
        pass

    @property
    @abstractmethod
    def GuideRateDeclination(self):
        pass

    @GuideRateDeclination.setter
    @abstractmethod
    def GuideRateDeclination(self, value):
        pass

    @property
    @abstractmethod
    def GuideRateRightAscension(self):
        pass

    @GuideRateRightAscension.setter
    @abstractmethod
    def GuideRateRightAscension(self, value):
        pass

    @property
    @abstractmethod
    def IsPulseGuiding(self):
        pass

    @property
    @abstractmethod
    def RightAscension(self):
        pass

    @property
    @abstractmethod
    def RightAscensionRate(self):
        pass

    @RightAscensionRate.setter
    @abstractmethod
    def RightAscensionRate(self, value):
        pass

    @property
    @abstractmethod
    def SideOfPier(self):
        pass

    @SideOfPier.setter
    @abstractmethod
    def SideOfPier(self, value):
        pass

    @property
    @abstractmethod
    def SiderealTime(self):
        pass

    @property
    @abstractmethod
    def SiteElevation(self):
        pass

    @SiteElevation.setter
    @abstractmethod
    def SiteElevation(self, value):
        pass

    @property
    @abstractmethod
    def SiteLatitude(self):
        pass

    @SiteLatitude.setter
    @abstractmethod
    def SiteLatitude(self, value):
        pass

    @property
    @abstractmethod
    def SiteLongitude(self):
        pass

    @property
    @abstractmethod
    def Slewing(self):
        pass

    @property
    @abstractmethod
    def SlewSettleTime(self):
        pass

    @SlewSettleTime.setter
    @abstractmethod
    def SlewSettleTime(self, value):
        pass

    @property
    @abstractmethod
    def TargetDeclination(self):
        pass

    @TargetDeclination.setter
    @abstractmethod
    def TargetDeclination(self, value):
        pass

    @property
    @abstractmethod
    def TargetRightAscension(self):
        pass

    @TargetRightAscension.setter
    @abstractmethod
    def TargetRightAscension(self, value):
        pass

    @property
    @abstractmethod
    def Tracking(self):
        pass

    @Tracking.setter
    @abstractmethod
    def Tracking(self, value):
        pass

    @property
    @abstractmethod
    def TrackingRate(self):
        pass

    @TrackingRate.setter
    @abstractmethod
    def TrackingRate(self, value):
        pass

    @property
    @abstractmethod
    def TrackingRates(self):
        pass

    @property
    @abstractmethod
    def UTCDate(self):
        pass

    @UTCDate.setter
    @abstractmethod
    def UTCDate(self, value):
        pass



"""telrun test docstring

This is a test docstring for the telrun module."""

# isort: skip_file

import logging

logger = logging.getLogger(__name__)

from .telrun_exception import TelrunException
from .exoplanet_transits import exoplanet_transits
from .init_dirs import init_telrun_dir, init_remote_dir
from .mk_mosaic_schedule import mk_mosaic_schedule
from .rst import rst
from .schedtel import schedtel, plot_schedule_gantt, parse_sch_file
from .startup import start_telrun, start_syncfiles
from .summary_report import summary_report
from .syncfiles import syncfiles
from .telrun_operator import TelrunOperator

__all__ = [
    "exoplanet_transits",
    "init_telrun_dir",
    "init_remote_dir",
    "mk_mosaic_schedule",
    "rst",
    "schedtel",
    "plot_schedule_gantt",
    "parse_sch_file",
    "start_telrun",
    "start_syncfiles",
    "summary_report",
    "syncfiles",
    "TelrunOperator",
    "TelrunException",
]



import configparser
import datetime
import logging

import click
import prettytable
import timezonefinder
from astropy import coordinates as coord
from astropy import time as astrotime
from astropy import units as u
from astroquery.ipac.nexsci.nasa_exoplanet_archive import NasaExoplanetArchive

from ..observatory import Observatory

logger = logging.getLogger(__name__)


@click.command(
    epilog="""Check out the documentation at
               https://pyscope.readthedocs.io/en/latest/
               for more information."""
)
@click.option(
    "-f",
    "--filt",
    type=str,
    default="g",
    show_default=True,
    help="SDSS filter magnitude to check",
)
@click.option(
    "-m",
    "--max-mag",
    type=int,
    default=15,
    show_default=True,
    help="Maximum magnitude of host star",
)
@click.option(
    "-t",
    "--transit-depth",
    type=float,
    default=0.02,
    show_default=True,
    help="Minimum transit depth",
)
@click.option(
    "-p",
    "--transit-depth-percent",
    type=float,
    help="Minimum transit depth in percent. Overrides -t option.",
)
@click.option(
    "-d",
    "--date",
    type=str,
    help="""Date of observation in YYYY-MM-DD format.
              If not specified, defaults to today.""",
)
@click.option(
    "-o",
    "--observatory",
    type=click.Path(exists=True, dir_okay=False, resolve_path=True),
    default="./config/observatory.cfg",
    help="Path to observatory config file",
)
@click.option(
    "-v",
    "--verbose",
    count=True,
    type=click.IntRange(0, 1),  # Range can be changed
    help="Increase verbosity",
)
@click.version_option()
def exoplanet_transits_cli(
    filt, max_mag, transit_depth, transit_depth_percent, date, observatory, verbose
):
    """

    Acknowledgements: This code was adapted from Aurora Hiveley's
    exoplanet_transits.py script.
    """

    logger.setLevel(int(10 * (1 - verbose)))  # Change range via 3
    logger.debug(f"Verbosity level set to {verbose}")
    logger.debug(
        f"""exoplanet_transits_cli(max_mag={max_mag},
                transit_depth={transit_depth},
                transit_depth_percent={transit_depth_percent},
                date={date},
                observatory={observatory},
                verbose={verbose})"""
    )

    if type(observatory) is str:
        observatory = Observatory(config_path=observatory)
    elif type(observatory) is Observatory:
        pass
    else:
        raise TypeError(
            "Observatory must be, a string, Observatory object, or astroplan.Observer object."
        )

    if transit_depth_percent is not None:
        transit_depth = np.log10(1 + transit_depth_percent / 100) / 0.4

    if date is None:
        logger.info("Using current date at observatory location")
        tz = timezonefinder.TimezoneFinder().timezone_at(
            lng=observatory.observatory_location.lon.deg,
            lat=observatory.observatory_location.lat.deg,
        )
        date = datetime.datetime.now(pytz.timezone(tz))

    t0 = (
        astrotime.Time(
            datetime.datetime(date.year, date.month, date.day, 12, 0, 0),
            format="datetime",
            scale="utc",
        )
        - (
            (
                astrotime.Time(date, format="datetime", scale="utc")
                - astrotime.Time.now()
            ).day
            % 1
        )
        * u.day
    )
    t1 = t0 + 1 * u.day
    logger.info("Searching UT range: %s to %s" % (t0.iso, t1.iso))

    min_ra = observatory.lst(t=t0).to(u.deg).value
    max_ra = observatory.lst(t=t1).to(u.deg).value
    logger.info("Searching LST range: %s to %s" % (min_ra, max_ra))

    min_dec = np.max(
        observatory.observatory_location.lat.deg - (90 - observatory.min_altitude), -90
    )
    max_dec = np.min(
        observatory.observatory_location.lat.deg + (90 - observatory.min_altitude), 90
    )
    logger.info("Searching declination range: %s to %s" % (min_dec, max_dec))

    transit_depth_percent = 100 * (10 ** (0.4 * transit_depth) - 1)

    table = NasaExoplanetArchive.query_criteria(
        table="PS",
        select=f"""pl_name,hostname,
                ra,dec,
                sy_{filt}mag,
                pl_tranmid,
                pl_orbper,
                pl_trandep,
                pl_trandur,
                rowupdate,
                st_nphot
                """,
        where=f"""
                pl_controv_flag=0 and
                tran_flag=1 and
                ra between {min_ra} and {max_ra} and
                dec between {min_dec} and {max_dec} and
                sy_{filt}mag <= {max_mag} and
                pl_trandep >= {transit_depth_percent}
                """,
        order=f"pl_trandep desc",
    )

    print_table = []

    for exo, i in enumerate(table):
        t_until_next_transit = exo["pl_orbper"] * u.day - (t0 - exo["pl_tranmid"])

        if t_until_next_transit > (t1 - t0):
            logger.info(
                f"Removing {exo['pl_name']} because it is not transiting between {t0.iso} and {t1.iso}"
            )
            t.remove_row(i)
            continue

        print_table.append(
            [
                exo["pl_name"],
                exo["hostname"],
                exo["sky_coord"].ra.hms,
                exo["sky_coord"].dec.dms,
                exo[f"sy_{filt}mag"],
                exo["pl_tranmid"],
                exo["pl_orbper"],
                exo["pl_trandep"],
                exo["pl_trandur"],
                exo["rowupdate"],
                exo["st_nphot"],
            ]
        )

    table = prettytable.PrettyTable()
    table.add_rows(print_table)
    table.set_style(prettytable.SINGLE_BORDER)
    table.field_names = [
        "Planet Name",
        "Host Name",
        "RA",
        "Dec",
        f"{filt} mag",
        "Reference Epoch [JD]",
        "Orbital Period [d]",
        f"Transit Depth [%]",
        "Transit Duration [h]",
        "Last Updated",
        "# of Photometric Observations",
    ]

    table.align["Planet Name"] = "l"
    click.echo(table)
    click.echo()
    clic, echo(f"Number of exoplanets = {len(print_table)}")

    return table


exoplanet_transits = exoplanet_transits_cli.callback



import logging

from .ascom_device import ASCOMDevice
from .safety_monitor import SafetyMonitor

logger = logging.getLogger(__name__)


class ASCOMSafetyMonitor(ASCOMDevice, SafetyMonitor):
    def __init__(self, identifier, alpaca=False, device_number=0, protocol="http"):
        super().__init__(
            identifier,
            alpaca=alpaca,
            device_type=self.__class__.__name__.replace("ASCOM", ""),
            device_number=device_number,
            protocol=protocol,
        )

    @property
    def IsSafe(self):
        logger.debug(f"ASCOMSafetyMonitor.IsSafe property called")
        return self._device.IsSafe



import logging
import platform
import time

logger = logging.getLogger(__name__)

from ..utils import _force_async
from .wcs import WCS


class PinpointWCS(WCS):
    def __init__(self):
        logger.debug("PinpointWCS.__init__() called")
        if platform.system() != "Windows":
            raise Exception("PinPoint is only available on Windows.")
        else:
            from win32com.client import Dispatch

            self._solver = Dispatch("PinPoint.Plate")

    def Solve(
        self,
        filepath,
        scale_est,
        ra_key="RA",
        dec_key="DEC",
        ra=None,
        dec=None,
        ra_dec_units=("hour", "deg"),
        solve_timeout=60,
        catalog=3,
        catalog_path="C:\GSC-1.1",
    ):
        logger.debug(
            f"""PinpointWCS.Solve(
            {filepath}, {scale_est}, {ra_key}, {dec_key}, {ra}, {dec},
            {ra_dec_units}, {solve_timeout}, {catalog}, {catalog_path}
        ) called"""
        )

        self._solver.AttachFITS(filepath)

        if kwargs.get("ra", None) is None or kwargs.get("dec", None) is None:
            with pyfits.open(filepath) as hdul:
                ra = hdul[0].header[ra_key]
                dec = hdul[0].header[dec_key]
        else:
            ra = kwargs.get("ra", self._solver.TargetRightAscension)
            dec = kwargs.get("dec", self._solver.TargetDeclination)

        obj = coord.SkyCoord(ra, dec, unit=ra_dec_units, frame="icrs")
        self._solver.RightAscension = obj.ra.hour
        self._solver.Declination = obj.dec.deg

        self._solver.ArcsecPerPixelHoriz = scale_est
        self._solver.ArcsecPerPixelVert = scale_est
        self.Catalog = catalog
        self.CatalogPath = catalog_path

        solved = None
        solved = self._async_solve()
        start_solve = time.time()
        while time.time() - start_solve < solve_timeout and solved is None:
            time.sleep(1)
        else:
            if solved is None:
                raise Exception("Pinpoint solve timed out.")
            else:
                return solved

        self._solver.UpdateFITS()
        self._solver.DetachFITS()
        return True

    @_force_async
    def _async_solve(self):
        self.Solver.Solve()



from abc import ABC, abstractmethod

from ._docstring_inheritee import _DocstringInheritee


class WCS(ABC, metaclass=_DocstringInheritee):
    @abstractmethod
    def __init__(self, *args, **kwargs):
        pass

    @abstractmethod
    def Solve(self, *args, **kwargs):
        pass



import configparser
import datetime
import logging
import os
import zoneinfo

import astroplan
import click
import numpy as np
import prettytable
import timezonefinder
from astropy import coordinates as coord
from astropy import table
from astropy import time as astrotime
from astropy import units as u
from astroquery.mpc import MPC

from ..observatory import Observatory

logger = logging.getLogger(__name__)


@click.command(
    epilog="""Check out the documentation at
               https://pyscope.readthedocs.io/en/latest/
               for more information."""
)
@click.option(
    "-s",
    "--source",
    type=str,
    help="""Source name.
    Is resolved by astropy.coordinates.SkyCoord.from_name() first, then
    if not found, attempts to check the Minor Planet Center via astroquery.
    Finally, it attempts to use astropy.coordinates.get_body for solar system
    bodies. If no source is given, then dawn and dusk times are calculated for the
    observatory.""",
)
@click.option(
    "-d",
    "--date",
    type=str,
    help="""Date of observation [YYYY-MM-DD].
    If none is given, the current date at the observatory is used.""",
)
@click.option(
    "-o",
    "--observatory",
    type=click.Path(resolve_path=True, dir_okay=False, exists=True),
    default="./config/observatory.cfg",
    show_default=True,
    help="""Observatory configuration file.""",
)
@click.option(
    "-v",
    "--verbose",
    count=True,
    type=click.IntRange(0, 1),  # Range can be changed
    help="Increase verbosity",
)
@click.version_option()
def rst_cli(source=None, date=None, observatory="./config/observatory.cfg", verbose=-1):
    """Calculate the rise, set, and transit times for a given source or dusk and
    dawn times for the observatory.\b

    The observatory configuration file is required to calculate the rise, set, and
    transit times for a given source. If no source is given, then the dusk and dawn
    times are calculated for the observatory. If no date is given, then the current
    date at the location of the observatory is used. The source is first resolved
    by `astropy.coordinates.SkyCoord.from_name` and if not found, the `astroquery.mpc.MPCClass`
    is used. Finally, if the source is not found, then `astropy.coordinates.get_body` is used
    for solar system bodies. If the source is still not found, then the program exits with an
    error message.

    .. warning::
        The rise, set, and transit times may be inaccurate for high proper motion
        bodies and solar system bodies. Other sources are generally accurate to
        within a few minutes.

    Parameters
    ----------
    source : str, optional
        Source name.

    date : str [YYYY-MM-DD], optional
        Date of observation. If none is given, the current date at the observatory is used.

    observatory : str or `~pyscope.observatory.Observatory`, default='./config/observatory.cfg'
        Observatory configuration file or :py:class:`~pyscope.observatory.Observatory` object.

    verbose : int, {-1, 0, 1}, default=-1
        Verbosity level. 0: no debug messages, 1: debug messages. Default set to -1
        for API use and 0 for CLI use.

    Returns
    -------
    events: list
        List of tuples containing the event name and a `~astropy.time.Time` object.

    Raises
    ------
    `click.BadParameter`

    See Also
    --------
    pyscope.telrun.exoplanet_transits
    pyscope.telrun.schedtel
    astropy.coordinates.SkyCoord.from_name
    astroquery.mpc.MPCClass.get_ephemeris

    Examples
    --------
    Compute the rise, set, and transit times for M64 on 2024-01-30.

    .. doctest::

        >>> from pyscope.telrun import rst
        >>> tbl = rst(observatory="../tests/observatory.cfg", source="M64", date="2024-01-30")
        Times from: 2024-01-30 12:00 (MST), 2024-01-30 19:00 (UTC)
        +-----------------------------+-------+-------+-------+
        |                        Time | Local |  UTC  |  LST  |
        +-----------------------------+-------+-------+-------+
        |                      Sunset | 17:50 | 00:50 |  2:07 |
        |        Civil Twilight Start | 18:20 | 01:20 |  2:37 |
        |     Nautical Twilight Start | 18:49 | 01:49 |  3:06 |
        | Astronomical Twilight Start | 19:18 | 02:18 |  3:35 |
        |                    M64 Rise | 21:44 | 04:44 |  6:02 |
        |                 M64 Transit | 04:39 | 11:39 | 12:58 |
        |   Astronomical Twilight End | 05:52 | 12:52 | 14:11 |
        |       Nautical Twilight End | 06:21 | 13:21 | 14:40 |
        |          Civil Twilight End | 06:50 | 13:50 | 15:09 |
        |                     Sunrise | 07:20 | 14:20 | 15:39 |
        |                     M64 Set | 11:34 | 18:34 | 19:54 |
        +-----------------------------+-------+-------+-------+

    Perform the same calculation, but on the command line.

    .. code-block:: shell

        $ rst -o ../tests/observatory.cfg -s M64 -d 2024-01-30
        Times from: 2024-01-30 12:00 (MST), 2024-01-30 19:00 (UTC)
        +-----------------------------+-------+-------+-------+
        |                        Time | Local |  UTC  |  LST  |
        +-----------------------------+-------+-------+-------+
        |                      Sunset | 17:50 | 00:50 |  2:07 |
        |        Civil Twilight Start | 18:20 | 01:20 |  2:37 |
        |     Nautical Twilight Start | 18:49 | 01:49 |  3:06 |
        | Astronomical Twilight Start | 19:18 | 02:18 |  3:35 |
        |                    M64 Rise | 21:44 | 04:44 |  6:02 |
        |                 M64 Transit | 04:39 | 11:39 | 12:58 |
        |   Astronomical Twilight End | 05:52 | 12:52 | 14:11 |
        |       Nautical Twilight End | 06:21 | 13:21 | 14:40 |
        |          Civil Twilight End | 06:50 | 13:50 | 15:09 |
        |                     Sunrise | 07:20 | 14:20 | 15:39 |
        |                     M64 Set | 11:34 | 18:34 | 19:54 |
        +-----------------------------+-------+-------+-------+

    """

    if verbose > -1:
        logger.setLevel(int(10 * (2 - verbose)))
        logger.addHandler(logging.StreamHandler())
    logger.debug(f"Verbosity level set to {verbose}")
    logger.debug(
        f"""rst(source={source}, date={date}, observatory={observatory}
        verbose={verbose})"""
    )

    if type(observatory) is str:
        observatory = os.path.abspath(observatory)
        if os.path.exists(observatory):
            config = configparser.ConfigParser()
            config.read(observatory)
            lat = coord.Latitude(config["site"]["latitude"], unit=u.deg)
            lon = coord.Longitude(
                config["site"]["longitude"], unit=u.deg, wrap_angle=180 * u.deg
            )
            name = config.get("site", "name", fallback="pyscope observatory")
        else:
            raise click.BadParameter(
                "Observatory configuration file does not exist: %s" % observatory
            )
    elif type(observatory) is Observatory:
        lat = observatory.observatory_location.lat
        lon = observatory.observatory_location.lon
        name = observatory.instrument_name
    else:
        raise click.BadParameter("Observatory must be a string or Observatory object.")
    logger.debug(f"lat = {lat}")
    logger.debug(f"lon = {lon}")
    logger.debug(f"name = {name}")

    tz = timezonefinder.TimezoneFinder().timezone_at(lng=lon.deg, lat=lat.deg)
    tz = zoneinfo.ZoneInfo(tz)
    logger.debug(f"tz = {tz}")

    if date is None:
        logger.debug("Using current date at observatory location")
        date = datetime.datetime.now()
    else:
        date = datetime.datetime.strptime(date, "%Y-%m-%d")
    date = datetime.datetime(date.year, date.month, date.day, 12, 0, 0, tzinfo=tz)

    t0 = astrotime.Time(
        datetime.datetime(date.year, date.month, date.day, 12, 0, 0, tzinfo=tz),
        format="datetime",
    )
    logger.debug(f"t0 = {t0}")

    observer = astroplan.Observer(
        location=coord.EarthLocation.from_geodetic(lon, lat), name=name
    )

    horizon = (
        ("Sunset", observer.sun_set_time(t0, which="nearest")),
        ("Sunrise", observer.sun_rise_time(t0, which="next")),
    )
    civil_twilight = (
        (
            "Civil Twilight Start",
            observer.sun_set_time(t0, which="nearest", horizon=-6 * u.deg),
        ),
        (
            "Civil Twilight End",
            observer.sun_rise_time(t0, which="next", horizon=-6 * u.deg),
        ),
    )
    nautical_twilight = (
        (
            "Nautical Twilight Start",
            observer.sun_set_time(t0, which="nearest", horizon=-12 * u.deg),
        ),
        (
            "Nautical Twilight End",
            observer.sun_rise_time(t0, which="next", horizon=-12 * u.deg),
        ),
    )
    astronomical_twilight = (
        (
            "Astronomical Twilight Start",
            observer.sun_set_time(t0, which="nearest", horizon=-18 * u.deg),
        ),
        (
            "Astronomical Twilight End",
            observer.sun_rise_time(t0, which="next", horizon=-18 * u.deg),
        ),
    )

    times = [
        horizon[0],
        horizon[1],
        civil_twilight[0],
        civil_twilight[1],
        nautical_twilight[0],
        nautical_twilight[1],
        astronomical_twilight[0],
        astronomical_twilight[1],
    ]

    if source is not None:
        try:
            obj = coord.SkyCoord.from_name(source)
        except coord.name_resolve.NameResolveError:
            logger.debug(f"Could not resolve {source} by name. Attempting MPC query.")
            try:
                eph = MPC.get_ephemeris(
                    source,
                    start=t0 + 0.5 * u.day,
                    location=observer.location,
                    proper_motion="sky",
                )
                obj = coord.SkyCoord(
                    ra=eph["RA"],
                    dec=eph["Dec"],
                    unit=("deg", "deg"),
                    pm_ra_cosdec=eph["dRA cos(Dec)"],
                    pm_dec=eph["dDec"],
                    frame="icrs",
                )
                logger.warning(
                    f"""Could not resolve {source} by name. Using MPC query.
                    Note that rise-set times may be inaccurate for high PM bodies."""
                )
            except:
                try:
                    obj = coord.get_body(source, t0 + 0.5 * u.day, observer.location)
                    logger.warning(
                        f"""Could not resolve {source} by name or MPC query. Using get_body query.
                        Note that rise-set times may be inaccurate for solar system bodies."""
                    )
                except:
                    click.echo(
                        f"Could not resolve {source} by name, MPC query, or get_body query."
                    )
                    click.Abort()
                    return
        logger.debug(f"obj = {obj}")

        rise_time = (
            "%s Rise" % source,
            observer.target_rise_time(t0, obj, which="next"),
        )
        transit_time = (
            "%s Transit" % source,
            observer.target_meridian_transit_time(t0, obj, which="next"),
        )
        set_time = ("%s Set" % source, observer.target_set_time(t0, obj, which="next"))

        logger.debug(f"rise_time = {rise_time}")
        logger.debug(f"transit_time = {transit_time}")
        logger.debug(f"set_time = {set_time}")

        logger.debug(type(transit_time[1].value))

        if not isinstance(rise_time[1].value, np.ma.MaskedArray):
            times.append(rise_time)
        if not isinstance(transit_time[1].value, np.ma.MaskedArray):
            times.append(transit_time)
        if not isinstance(set_time[1].value, np.ma.MaskedArray):
            times.append(set_time)

        if isinstance(rise_time[1].value, np.ma.MaskedArray) and isinstance(
            set_time[1].value, np.ma.MaskedArray
        ):
            if isinstance(transit_time[1].value, np.ma.MaskedArray):
                logger.warning("This source never rises above the horizon.")
            else:
                logger.warning("This source is circumpolar.")

    times.sort(key=lambda x: x[1].jd)

    tbl = prettytable.PrettyTable()
    tbl.field_names = ["Time", "Local", "UTC", "LST"]
    tbl.align["Time"] = "r"

    for t in times:
        tbl.add_row(
            [
                t[0],
                t[1].to_datetime(timezone=tz).strftime("%H:%M"),
                t[1].strftime("%H:%M"),
                observer.local_sidereal_time(t[1]).to_string(
                    precision=0, sep=":", fields=2
                ),
            ]
        )

    if verbose > -1:
        click.echo(
            f"Times from: {date.strftime('%Y-%m-%d %H:%M (%Z)')}, {t0.strftime('%Y-%m-%d %H:%M (UTC)')}"
        )
        click.echo(tbl)

    return times


rst = rst_cli.callback



import ast
import os
import platform
import sys
import time
import xml.etree.ElementTree as ET

import pytest

from pyscope import observatory

# Much of these tests are inspired by the tests in the alpyca repo
# found here: https://github.com/ASCOMInitiative/alpyca


def pytest_configure():
    pytest.simulator_server = observatory.SimulatorServer()
    time.sleep(12)


def pytest_sessionfinish(session, exitstatus):
    del pytest.simulator_server


@pytest.fixture()
def device(request):
    global d
    n = "ASCOM" + "".join(
        [s.capitalize() for s in request.module.__name__.split("_")[2:]]
    )
    c = getattr(observatory, n)  # Creates a device class by string name :-)
    d = c("localhost:32323", alpaca=True)  # Created an instance of the class
    #    d = c('[fe80::9927:65fc:e9e8:f33a%eth0]:32323', 0)  # RPi 4 Ethernet to Windows OmniSim IPv6
    d.Connected = True
    return d


@pytest.fixture()
def disconnect(request):
    global d
    yield
    d.Connected = False
    n = "ASCOM" + "".join(
        [s.capitalize() for s in request.module.__name__.split("_")[2:]]
    )


@pytest.fixture()
def winer_oc(request):
    global d
    d = observatory.HTMLObservingConditions(
        "https://winer.org/Site/Weather.php",
        rain_rate_keyword="RAIN",
        rain_rate_units="in",
        sky_brightness_keyword="BRIGHTNESS",
        sky_brightness_units="Vmag/sq.asec",
        star_fwhm_keyword="LAST_SEEING",
        star_fwhm_units='"',
        wind_direction_keyword="WINDDIR",
    )
    return d


@pytest.fixture()
def winer_sm(request):
    global d
    d = observatory.HTMLSafetyMonitor(
        "https://winer.org/Site/Roof.php",
        check_phrase=b"ROOFPOSITION=OPEN",
    )
    return d


"""@pytest.fixture()
def settings(request):
    n = request.module.__name__.split('_')[-1].capitalize()
    if platform.system() == "Windows":
        data_file = f"{os.getenv('USERPROFILE')}/.ASCOM/Alpaca/ASCOM-Alpaca-Simulator/{n}/v1/Instance-0.xml"
    else:                       # TODO No MacOS
        n = n.lower()
        data_file = f"{os.getenv('HOME')}/.config/ascom/alpaca/ascom-alpaca-simulator/{n}/v1/instance-0.xml"
    tree = ET.parse(data_file)
    root = tree.getroot()
    s = {}
    for i in root.iter("SettingsPair"):
        k = i.find('Key').text
        v = i.find('Value').text
        try:
            s[k] = ast.literal_eval(v)
        except:
            s[k] = v
    return s

def get_settings(device):
    if platform.system() == "Windows":
        data_file = f"{os.getenv('USERPROFILE')}/.ASCOM/Alpaca/ASCOM-Alpaca-Simulator/{device}/v1/Instance-0.xml"
    else:                       # TODO No MacOS
        device = device.lower()
        data_file = f"{os.getenv('HOME')}/.config/ascom/alpaca/ascom-alpaca-simulator/{device}/v1/instance-0.xml"
    tree = ET.parse(data_file)
    root = tree.getroot()
    s = {}
    for i in root.iter("SettingsPair"):
        k = i.find('Key').text
        v = i.find('Value').text
        try:
            s[k] = ast.literal_eval(v)
        except:
            s[k] = v
    return s"""



def _get_number_from_line(line, expected_keyword, expected_units, is_numeric):
    """
    Check to see if the provided line looks like a valid telemetry line
    from the Winer webpage. A typical line looks like this:
        <!-- TEMPERATURE=7None0 F -->

    If the line matches this format and the contents match expectations, return
    the extracted value from the line.

    line: the line text to inspect
    expected_keyword: the line must contain this keyword ('TEMPERATURE' in the example above)
    expected_units: the line must contain these units after the value ('F' in the example above).
        If this value is None, then units are not validated
    is_numeric: if True, the value is validated and converted to a float before being returned.
                if False, the string value is returned

    If the line does not match or there is a problem (e.g. converting the value to a float),
    the function returns None.

    Otherwise, the function returns the value, either as a float (if requested) or as a string
    """

    line = line.strip()
    if not line.startswith(b"<!--"):
        return None
    if not line.endswith(b"-->"):
        return None

    line = line[4:-3]  # Strip off beginning and ending comment characters

    fields = line.split(b"=", 1)  # Split into at most two fields (keyword and value)
    if len(fields) != 2:
        return None

    line_keyword = fields[0].strip()
    line_value_and_units = fields[1].strip()

    fields = line_value_and_units.split(b" ", 1)
    line_value = fields[0].strip()
    if len(fields) > 1:
        line_units = fields[1]
    else:
        line_units = ""

    if line_keyword != expected_keyword:
        return None
    if expected_units is not None and line_units != expected_units:
        return None
    if is_numeric:
        try:
            return float(line_value)
        except:
            return None
    else:
        return line_value



import logging

from .ascom_device import ASCOMDevice
from .telescope import Telescope

logger = logging.getLogger(__name__)


class ASCOMTelescope(ASCOMDevice, Telescope):
    def __init__(self, identifier, alpaca=False, device_number=0, protocol="http"):
        super().__init__(
            identifier,
            alpaca=alpaca,
            device_type=self.__class__.__name__.replace("ASCOM", ""),
            device_number=device_number,
            protocol=protocol,
        )

    def AbortSlew(self):
        logger.debug("ASCOMTelescope.AbortSlew() called")
        self._device.AbortSlew()

    def AxisRates(self, Axis):
        logger.debug(f"ASCOMTelescope.AxisRates({Axis}) called")
        return self._device.AxisRates(Axis)

    def CanMoveAxis(self, Axis):
        logger.debug(f"ASCOMTelescope.CanMoveAxis({Axis}) called")
        return self._device.CanMoveAxis(Axis)

    def DestinationSideOfPier(self, RightAscension, Declination):
        logger.debug(
            f"ASCOMTelescope.DestinationSideOfPier({RightAscension}, {Declination}) called"
        )
        return self._device.DestinationSideOfPier(RightAscension, Declination)

    def FindHome(self):
        logger.debug("ASCOMTelescope.FindHome() called")
        self._device.FindHome()

    def MoveAxis(self, Axis, Rate):
        logger.debug(f"ASCOMTelescope.MoveAxis({Axis}, {Rate}) called")
        self._device.MoveAxis(Axis, Rate)

    def Park(self):
        logger.debug("ASCOMTelescope.Park() called")
        self._device.Park()

    def PulseGuide(self, Direction, Duration):
        logger.debug(f"ASCOMTelescope.PulseGuide({Direction}, {Duration}) called")
        self._device.PulseGuide(Direction, Duration)

    def SetPark(self):
        logger.debug("ASCOMTelescope.SetPark() called")
        self._device.SetPark()

    def SlewToAltAz(self, Azimuth, Altitude):  # pragma: no cover
        """
        .. deprecated:: 0.1.1
            ASCOM is deprecating this method.
        """
        logger.debug(f"ASCOMTelescope.SlewToAltAz({Azimuth}, {Altitude}) called")
        self._device.SlewToAltAz(Azimuth, Altitude)

    def SlewToAltAzAsync(self, Azimuth, Altitude):
        logger.debug(f"ASCOMTelescope.SlewToAltAzAsync({Azimuth}, {Altitude}) called")
        self._device.SlewToAltAzAsync(Azimuth, Altitude)

    def SlewToCoordinates(self, RightAscension, Declination):  # pragma: no cover
        """
        .. deprecated:: 0.1.1
            ASCOM is deprecating this method.
        """
        logger.debug(
            f"ASCOMTelescope.SlewToCoordinates({RightAscension}, {Declination}) called"
        )
        self._device.SlewToCoordinates(RightAscension, Declination)

    def SlewToCoordinatesAsync(self, RightAscension, Declination):
        logger.debug(
            f"ASCOMTelescope.SlewToCoordinatesAsync({RightAscension}, {Declination}) called"
        )
        self._device.SlewToCoordinatesAsync(RightAscension, Declination)

    def SlewToTarget(self):  # pragma: no cover
        """
        .. deprecated:: 0.1.1
            ASCOM is deprecating this method.
        """
        logger.debug("ASCOMTelescope.SlewToTarget() called")
        self._device.SlewToTarget()

    def SlewToTargetAsync(self):
        logger.debug("ASCOMTelescope.SlewToTargetAsync() called")
        self._device.SlewToTargetAsync()

    def SyncToAltAz(self, Azimuth, Altitude):
        logger.debug(f"ASCOMTelescope.SyncToAltAz({Azimuth}, {Altitude}) called")
        self._device.SyncToAltAz(Azimuth, Altitude)

    def SyncToCoordinates(self, RightAscension, Declination):
        logger.debug(
            f"ASCOMTelescope.SyncToCoordinates({RightAscension}, {Declination}) called"
        )
        self._device.SyncToCoordinates(RightAscension, Declination)

    def SyncToTarget(self):
        logger.debug("ASCOMTelescope.SyncToTarget() called")
        self._device.SyncToTarget()

    def Unpark(self):
        logger.debug("ASCOMTelescope.Unpark() called")
        self._device.Unpark()

    @property
    def AlignmentMode(self):
        logger.debug("ASCOMTelescope.AlignmentMode property accessed")
        return self._device.AlignmentMode

    @property
    def Altitude(self):
        logger.debug("ASCOMTelescope.Altitude property accessed")
        return self._device.Altitude

    @property
    def ApertureArea(self):
        logger.debug("ASCOMTelescope.ApertureArea property accessed")
        return self._device.ApertureArea

    @property
    def ApertureDiameter(self):
        logger.debug("ASCOMTelescope.ApertureDiameter property accessed")
        return self._device.ApertureDiameter

    @property
    def AtHome(self):
        logger.debug("ASCOMTelescope.AtHome property accessed")
        return self._device.AtHome

    @property
    def AtPark(self):
        logger.debug("ASCOMTelescope.AtPark property accessed")
        return self._device.AtPark

    @property
    def Azimuth(self):
        logger.debug("ASCOMTelescope.Azimuth property accessed")
        return self._device.Azimuth

    @property
    def CanFindHome(self):
        logger.debug("ASCOMTelescope.CanFindHome property accessed")
        return self._device.CanFindHome

    @property
    def CanPark(self):
        logger.debug("ASCOMTelescope.CanPark property accessed")
        return self._device.CanPark

    def CanPulseGuide(self, Direction):
        logger.debug(f"ASCOMTelescope.CanPulseGuide({Direction}) called")
        return self._device.CanPulseGuide(Direction)

    @property
    def CanSetDeclinationRate(self):
        logger.debug("ASCOMTelescope.CanSetDeclinationRate property accessed")
        return self._device.CanSetDeclinationRate

    @property
    def CanSetGuideRates(self):
        logger.debug("ASCOMTelescope.CanSetGuideRates property accessed")
        return self._device.CanSetGuideRates

    @property
    def CanSetPark(self):
        logger.debug("ASCOMTelescope.CanSetPark property accessed")
        return self._device.CanSetPark

    @property
    def CanSetPierSide(self):
        logger.debug("ASCOMTelescope.CanSetPierSide property accessed")
        return self._device.CanSetPierSide

    @property
    def CanSetRightAscensionRate(self):
        logger.debug("ASCOMTelescope.CanSetRightAscensionRate property accessed")
        return self._device.CanSetRightAscensionRate

    @property
    def CanSetTracking(self):
        logger.debug("ASCOMTelescope.CanSetTracking property accessed")
        return self._device.CanSetTracking

    @property
    def CanSlew(self):  # pragma: no cover
        """
        .. deprecated:: 0.1.1
            ASCOM is deprecating this property.
        """
        logger.debug("ASCOMTelescope.CanSlew property accessed")
        return self._device.CanSlew

    @property
    def CanSlewAltAz(self):  # pragma: no cover
        """
        .. deprecated:: 0.1.1
            ASCOM is deprecating this property.
        """
        logger.debug("ASCOMTelescope.CanSlewAltAz property accessed")
        return self._device.CanSlewAltAz

    @property
    def CanSlewAltAzAsync(self):
        logger.debug("ASCOMTelescope.CanSlewAltAzAsync property accessed")
        return self._device.CanSlewAltAzAsync

    @property
    def CanSlewAsync(self):
        logger.debug("ASCOMTelescope.CanSlewAsync property accessed")
        return self._device.CanSlewAsync

    @property
    def CanSync(self):
        logger.debug("ASCOMTelescope.CanSync property accessed")
        return self._device.CanSync

    @property
    def CanSyncAltAz(self):
        logger.debug("ASCOMTelescope.CanSyncAltAz property accessed")
        return self._device.CanSyncAltAz

    @property
    def CanUnpark(self):
        logger.debug("ASCOMTelescope.CanUnpark property accessed")
        return self._device.CanUnpark

    @property
    def Declination(self):
        logger.debug("ASCOMTelescope.Declination property accessed")
        return self._device.Declination

    @property
    def DeclinationRate(self):
        logger.debug("ASCOMTelescope.DeclinationRate property accessed")
        return self._device.DeclinationRate

    @DeclinationRate.setter
    def DeclinationRate(self, value):
        logger.debug(f"ASCOMTelescope.DeclinationRate set to {value}")
        self._device.DeclinationRate = value

    @property
    def DoesRefraction(self):
        logger.debug("ASCOMTelescope.DoesRefraction property accessed")
        return self._device.DoesRefraction

    @DoesRefraction.setter
    def DoesRefraction(self, value):
        logger.debug(f"ASCOMTelescope.DoesRefraction set to {value}")
        self._device.DoesRefraction = value

    @property
    def EquatorialSystem(self):
        logger.debug("ASCOMTelescope.EquatorialSystem property accessed")
        return self._device.EquatorialSystem

    @property
    def FocalLength(self):
        logger.debug("ASCOMTelescope.FocalLength property accessed")
        return self._device.FocalLength

    @property
    def GuideRateDeclination(self):
        logger.debug("ASCOMTelescope.GuideRateDeclination property accessed")
        return self._device.GuideRateDeclination

    @GuideRateDeclination.setter
    def GuideRateDeclination(self, value):
        logger.debug(f"ASCOMTelescope.GuideRateDeclination set to {value}")
        self._device.GuideRateDeclination = value

    @property
    def GuideRateRightAscension(self):
        logger.debug("ASCOMTelescope.GuideRateRightAscension property accessed")
        return self._device.GuideRateRightAscension

    @GuideRateRightAscension.setter
    def GuideRateRightAscension(self, value):
        logger.debug(f"ASCOMTelescope.GuideRateRightAscension set to {value}")
        self._device.GuideRateRightAscension = value

    @property
    def IsPulseGuiding(self):
        logger.debug("ASCOMTelescope.IsPulseGuiding property accessed")
        return self._device.IsPulseGuiding

    @property
    def RightAscension(self):
        logger.debug("ASCOMTelescope.RightAscension property accessed")
        return self._device.RightAscension

    @property
    def RightAscensionRate(self):
        logger.debug("ASCOMTelescope.RightAscensionRate property accessed")
        return self._device.RightAscensionRate

    @RightAscensionRate.setter
    def RightAscensionRate(self, value):
        logger.debug(f"ASCOMTelescope.RightAscensionRate set to {value}")
        self._device.RightAscensionRate = value

    @property
    def SideOfPier(self):
        logger.debug("ASCOMTelescope.SideOfPier property accessed")
        return self._device.SideOfPier

    @SideOfPier.setter
    def SideOfPier(self, value):
        logger.debug(f"ASCOMTelescope.SideOfPier set to {value}")
        self._device.SideOfPier = value

    @property
    def SiderealTime(self):
        logger.debug("ASCOMTelescope.SiderealTime property accessed")
        return self._device.SiderealTime

    @property
    def SiteElevation(self):
        logger.debug("ASCOMTelescope.SiteElevation property accessed")
        return self._device.SiteElevation

    @SiteElevation.setter
    def SiteElevation(self, value):
        logger.debug(f"ASCOMTelescope.SiteElevation set to {value}")
        self._device.SiteElevation = value

    @property
    def SiteLatitude(self):
        logger.debug("ASCOMTelescope.SiteLatitude property accessed")
        return self._device.SiteLatitude

    @SiteLatitude.setter
    def SiteLatitude(self, value):
        logger.debug(f"ASCOMTelescope.SiteLatitude set to {value}")
        self._device.SiteLatitude = value

    @property
    def SiteLongitude(self):
        logger.debug("ASCOMTelescope.SiteLongitude property accessed")
        return self._device.SiteLongitude

    @SiteLongitude.setter
    def SiteLongitude(self, value):
        logger.debug(f"ASCOMTelescope.SiteLongitude set to {value}")
        self._device.SiteLongitude = value

    @property
    def Slewing(self):
        logger.debug("ASCOMTelescope.Slewing property accessed")
        return self._device.Slewing

    @property
    def SlewSettleTime(self):
        logger.debug("ASCOMTelescope.SlewSettleTime property accessed")
        return self._device.SlewSettleTime

    @SlewSettleTime.setter
    def SlewSettleTime(self, value):
        logger.debug(f"ASCOMTelescope.SlewSettleTime set to {value}")
        self._device.SlewSettleTime = value

    @property
    def TargetDeclination(self):
        logger.debug("ASCOMTelescope.TargetDeclination property accessed")
        return self._device.TargetDeclination

    @TargetDeclination.setter
    def TargetDeclination(self, value):
        logger.debug(f"ASCOMTelescope.TargetDeclination set to {value}")
        self._device.TargetDeclination = value

    @property
    def TargetRightAscension(self):
        logger.debug("ASCOMTelescope.TargetRightAscension property accessed")
        return self._device.TargetRightAscension

    @TargetRightAscension.setter
    def TargetRightAscension(self, value):
        logger.debug(f"ASCOMTelescope.TargetRightAscension set to {value}")
        self._device.TargetRightAscension = value

    @property
    def Tracking(self):
        logger.debug("ASCOMTelescope.Tracking property accessed")
        return self._device.Tracking

    @Tracking.setter
    def Tracking(self, value):
        logger.debug(f"ASCOMTelescope.Tracking set to {value}")
        self._device.Tracking = value

    @property
    def TrackingRate(self):
        logger.debug("ASCOMTelescope.TrackingRate property accessed")
        return self._device.TrackingRate

    @TrackingRate.setter
    def TrackingRate(self, value):
        logger.debug(f"ASCOMTelescope.TrackingRate set to {value}")
        self._device.TrackingRate = value

    @property
    def TrackingRates(self):
        logger.debug("ASCOMTelescope.TrackingRates property accessed")
        return self._device.TrackingRates

    @property
    def UTCDate(self):
        logger.debug("ASCOMTelescope.UTCDate property accessed")
        return self._device.UTCDate

    @UTCDate.setter
    def UTCDate(self, value):
        logger.debug(f"ASCOMTelescope.UTCDate set to {value}")
        self._device.UTCDate = value



import logging

from .ascom_device import ASCOMDevice
from .dome import Dome

logger = logging.getLogger(__name__)


class ASCOMDome(ASCOMDevice, Dome):
    def __init__(self, identifier, alpaca=False, device_number=0, protocol="http"):
        super().__init__(
            identifier,
            alpaca=alpaca,
            device_type=self.__class__.__name__.replace("ASCOM", ""),
            device_number=device_number,
            protocol=protocol,
        )

    def AbortSlew(self):
        logger.debug(f"ASCOMDome.AbortSlew() called")
        self._device.AbortSlew()

    def CloseShutter(self):
        logger.debug(f"ASCOMDome.CloseShutter() called")
        self._device.CloseShutter()

    def FindHome(self):
        logger.debug(f"ASCOMDome.FindHome() called")
        self._device.FindHome()

    def OpenShutter(self):
        logger.debug(f"ASCOMDome.OpenShutter() called")
        self._device.OpenShutter()

    def Park(self):
        logger.debug(f"ASCOMDome.Park() called")
        self._device.Park()

    def SetPark(self):
        logger.debug(f"ASCOMDome.SetPark() called")
        self._device.SetPark()

    def SlewToAltitude(self, Altitude):
        logger.debug(f"ASCOMDome.SlewToAltitude({Altitude}) called")
        self._device.SlewToAltitude(Altitude)

    def SlewToAzimuth(self, Azimuth):
        logger.debug(f"ASCOMDome.SlewToAzimuth({Azimuth}) called")
        self._device.SlewToAzimuth(Azimuth)

    def SyncToAzimuth(self, Azimuth):
        logger.debug(f"ASCOMDome.SyncToAzimuth({Azimuth}) called")
        self._device.SyncToAzimuth(Azimuth)

    @property
    def Altitude(self):
        logger.debug(f"ASCOMDome.Altitude property called")
        return self._device.Altitude

    @property
    def AtHome(self):
        logger.debug(f"ASCOMDome.AtHome property called")
        return self._device.AtHome

    @property
    def AtPark(self):
        logger.debug(f"ASCOMDome.AtPark property called")
        return self._device.AtPark

    @property
    def Azimuth(self):
        logger.debug(f"ASCOMDome.Azimuth property called")
        return self._device.Azimuth

    @property
    def CanFindHome(self):
        logger.debug(f"ASCOMDome.CanFindHome property called")
        return self._device.CanFindHome

    @property
    def CanPark(self):
        logger.debug(f"ASCOMDome.CanPark property called")
        return self._device.CanPark

    @property
    def CanSetAltitude(self):
        logger.debug(f"ASCOMDome.CanSetAltitude property called")
        return self._device.CanSetAltitude

    @property
    def CanSetAzimuth(self):
        logger.debug(f"ASCOMDome.CanSetAzimuth property called")
        return self._device.CanSetAzimuth

    @property
    def CanSetPark(self):
        logger.debug(f"ASCOMDome.CanSetPark property called")
        return self._device.CanSetPark

    @property
    def CanSetShutter(self):
        logger.debug(f"ASCOMDome.CanSetShutter property called")
        return self._device.CanSetShutter

    @property
    def CanSlave(self):
        logger.debug(f"ASCOMDome.CanSlave property called")
        return self._device.CanSlave

    @property
    def CanSyncAzimuth(self):
        logger.debug(f"ASCOMDome.CanSyncAzimuth property called")
        return self._device.CanSyncAzimuth

    @property
    def ShutterStatus(self):
        logger.debug(f"ASCOMDome.ShutterStatus property called")
        return self._device.ShutterStatus

    @property
    def Slaved(self):
        logger.debug(f"ASCOMDome.Slaved property called")
        return self._device.Slaved

    @property
    def Slewing(self):
        logger.debug(f"ASCOMDome.Slewing property called")
        return self._device.Slewing



import configparser
import datetime
import json
import logging
import os
import shlex

import astroplan
import click
import cmcrameri as ccm
import matplotlib.dates as mdates
import matplotlib.pyplot as plt
import pytz
import smplotlib
import timezonefinder
from astropy import coordinates as coord
from astropy import time as astrotime
from astropy import units as u
from astroquery import mpc
from matplotlib import ticker

from .. import utils
from ..observatory import Observatory

logger = logging.getLogger(__name__)


@click.command()
@click.option(
    "-c",
    "--catalog",
    type=click.Path(resolve_path=True),
    default="schedules/schedule.cat",
    show_default=True,
    help="A path to a .cat file containing a list of .sch files, a single .sch file, or a list of ObservingBlocks to be scheduled.",
)
@click.option(
    "-i",
    "--ignore-order",
    "ignore_order",
    is_flag=True,
    default=False,
    show_default=True,
    help="Ignore the order of the .sch files in the catalog, acting as if there is only one .sch file. By default, the .sch files are scheduled one at a time.",
)
@click.option(
    "-d",
    "--date",
    type=click.DateTime(formats=["%m-%d-%Y"]),
    default=None,
    show_default=False,
    help="The local date at the observatory of the night to be scheduled. By default, the current date at the observatory location is used.",
)
@click.option(
    "-o",
    "--observatory",
    type=click.Path(resolve_path=True),
    default="./config/observatory.cfg",
    show_default=True,
    help=("The configuration file of the observatory that will execute this schedule"),
)
@click.option(
    "-m",
    "--max-altitude",
    "max_altitude",
    type=click.FloatRange(max=90),
    default=-12,
    show_default=True,
    help="The maximum altitude of the Sun for the night [degrees]. Civil twilight is -6, nautical twilight is -12, and astronomical twilight is -18.",
)
@click.option(
    "-e",
    "--elevation",
    type=click.FloatRange(min=0, max=90, clamp=True),
    default=30,
    show_default=True,
    help='The minimum elevation of all targets [degrees]. This is a "boolean" constraint; that is, there is no understanding of where it is closer to ideal to observe the target. To implement a preference for higher elevation targets, the airmass constraint should be used.',
)
@click.option(
    "-a",
    "--airmass",
    type=click.FloatRange(min=1),
    default=3,
    show_default=True,
    help="The maximum airmass of all targets. If not specified, the airmass is not constrained.",
)
@click.option(
    "-c",
    "--moon-separation",
    type=click.FloatRange(min=0),
    default=30,
    show_default=True,
    help="The minimum angular separation between the Moon and all targets [degrees].",
)
@click.option(
    "-s",
    "--scheduler",
    nargs=2,
    type=(click.Path(exists=True, resolve_path=True), str),
    default=("", ""),
    show_default=False,
    help=(
        "The filepath to and name of an astroplan.Scheduler custom sub-class. By default, the astroplan.PriorityScheduler is used."
    ),
)
@click.option(
    "-g",
    "--gap-time",
    "gap_time",
    type=click.FloatRange(min=0),
    default=60,
    show_default=True,
    help="The maximumum length of time a transition between ObservingBlocks could take [seconds].",
)
@click.option(
    "-r",
    "--resolution",
    type=click.FloatRange(min=0),
    default=5,
    show_default=True,
    help="The time resolution of the schedule [seconds].",
)
@click.option(
    "-f",
    "--filename",
    type=click.Path(),
    default=None,
    show_default=False,
    help="The output file name. The file name is formatted with the UTC date of the first observation in the schedule. By default, it is placed in the current working directory, but if a path is specified, the file will be placed there. WARNING: If the file already exists, it will be overwritten.",
)
@click.option(
    "-t",
    "--telrun",
    is_flag=True,
    default=False,
    show_default=True,
    help="Places the output file in ./schedules/ or in the directory specified by the $TELRUN_EXECUTE environment variable. Causes -f/--filename to be ignored. WARNING: If the file already exists, it will be overwritten.",
)
@click.option(
    "-p",
    "--plot",
    type=click.IntRange(1, 3),
    default=None,
    show_default=False,
    help="Plots the schedule. 1: plots the schedule as a Gantt chart. 2: plots the schedule by target with airmass. 3: plots the schedule on a sky chart.",
)
@click.option(
    "-q", "--quiet", is_flag=True, default=False, show_default=True, help="Quiet output"
)
@click.option(
    "-v", "--verbose", default=0, show_default=True, count=True, help="Verbose output"
)
@click.version_option()
def schedtel_cli(
    catalog,
    ignore_order,
    date,
    observatory,
    max_altitude,
    elevation,
    airmass,
    moon_separation,
    scheduler,
    gap_time,
    resolution,
    filename,
    telrun,
    plot,
    quiet,
    verbose,
):
    # Set up logging
    if quiet:
        level = logging.ERROR
    elif verbose == 0:
        level = logging.WARNING
    elif verbose == 1:
        level = logging.INFO
    elif verbose >= 2:
        level = logging.DEBUG
    logging.basicConfig(level=level)

    logger.info("Starting schedtel")
    logger.debug(f"catalog: {catalog}")
    logger.debug(f"ignore_order: {ignore_order}")
    logger.debug(f"date: {date}")
    logger.debug(f"observatory: {observatory}")
    logger.debug(f"max_altitude: {max_altitude}")
    logger.debug(f"elevation: {elevation}")
    logger.debug(f"airmass: {airmass}")
    logger.debug(f"moon_separation: {moon_separation}")
    logger.debug(f"scheduler: {scheduler}")
    logger.debug(f"gap_time: {gap_time}")
    logger.debug(f"resolution: {resolution}")
    logger.debug(f"filename: {filename}")
    logger.debug(f"telrun: {telrun}")
    logger.debug(f"plot: {plot}")
    logger.debug(f"quiet: {quiet}")
    logger.debug(f"verbose: {verbose}")

    blocks = []

    if os.path.isfile(catalog):
        logger.debug(f"catalog is a file")
        if catalog.endswith(".cat"):
            with open(catalog, "r") as f:
                sch_files = f.read().splitlines()
            for f in sch_files:
                if not os.path.isfile(f):
                    logger.error(f"File {f} in catalog {catalog} does not exist.")
                    return
                blocks.append(parse_sch_file(f))
        elif catalog.endswith(".sch"):
            blocks.append(parse_sch_file(catalog))

    elif type(catalog) in (list, tuple, iter):
        logger.debug(f"catalog is a list")
        for block in catalog:
            if type(block) in (list, tuple, iter):
                for b in block:
                    if type(b) is not astroplan.ObservingBlock:
                        logger.error(
                            f"Object {b} in catalog {catalog} is not an astroplan.ObservingBlock."
                        )
                        return
                blocks.append(block)
            elif type(block) is astroplan.ObservingBlock:
                blocks.append([block])
            else:
                logger.error(
                    f"Object {block} in catalog {catalog} is not an astroplan.ObservingBlock."
                )
                return

    if ignore_order:
        blocks = [
            [blocks[i][j] for i in range(len(blocks)) for j in range(len(blocks[i]))]
        ]

    # Define the observatory
    logger.info("Parsing the observatory")
    if type(observatory) is not astroplan.Observer:
        if type(observatory) is str:
            obs_cfg = configparser.ConfigParser()
            obs_cfg.read(observatory)
            obs_long = obs_cfg.getfloat("site", "longitude")
            obs_lat = obs_cfg.getfloat("site", "latitude")
            obs_height = obs_cfg.getfloat("site", "elevation")
            slew_rate = obs_cfg.getfloat("scheduling", "slew_rate") * u.deg / u.second
            instrument_reconfiguration_times = json.loads(
                obs_cfg.get("scheduling", "instrument_reconfiguration_times")
            )
            instrument_name = obs_cfg.get("site", "instrument_name")
            observatory = astroplan.Observer(
                location=coord.EarthLocation(
                    lon=obs_long * u.deg, lat=obs_lat * u.deg, height=obs_height * u.m
                )
            )
        elif type(observatory) is Observatory:
            obs_long = observatory.observatory_location.lon.deg
            obs_lat = observatory.observatory_location.lat.deg
            obs_height = observatory.observatory_location.height.m
            slew_rate = observatory.slew_rate * u.deg / u.second
            instrument_reconfiguration_times = (
                observatory.instrument_reconfiguration_times
            )
            instrument_name = observatory.instrument_name
            observatory = astroplan.Observer(location=observatory.observatory_location)
        else:
            raise TypeError(
                "Observatory must be, a string, Observatory object, or astroplan.Observer object."
            )

    # Constraints
    logger.info("Defining global constraints")
    global_constraints = [
        astroplan.AltitudeConstraint(min=elevation * u.deg),
        astroplan.AtNightConstraint(max_solar_altitude=max_altitude * u.deg),
        astroplan.AirmassConstraint(max=airmass, boolean_constraint=False),
        astroplan.MoonSeparationConstraint(min=moon_separation * u.deg),
    ]

    # Transitioner
    logger.info("Defining transitioner")
    transitioner = astroplan.Transitioner(
        slew_rate, instrument_reconfiguration_times=instrument_reconfiguration_times
    )

    # Schedule
    if date is None:
        logger.info("Using current date at observatory location")
        tz = timezonefinder.TimezoneFinder().timezone_at(lng=obs_long, lat=obs_lat)
        date = datetime.datetime.now(pytz.timezone(tz))

    t0 = (
        astrotime.Time(
            datetime.datetime(date.year, date.month, date.day, 12, 0, 0),
            format="datetime",
            scale="utc",
        )
        - (
            (
                astrotime.Time(date, format="datetime", scale="utc")
                - astrotime.Time.now()
            ).day
            % 1
        )
        * u.day
    )
    t1 = t0 + 1 * u.day
    logger.info("Schedule time range: %s to %s" % (t0.iso, t1.iso))

    schedule = astroplan.Schedule(t0, t1)

    # Scheduler
    if scheduler[0] == "":
        logger.info("Using default scheduler: astroplan.PriorityScheduler")
        schedule_handler = astroplan.PriorityScheduler(
            constraints=global_constraints,
            observer=observatory,
            transitioner=transitioner,
            gap_time=gap_time * u.second,
            time_resolution=time_resolution * u.second,
        )
    else:
        logger.info(f"Using custom scheduler: {scheduler[0]}")
        spec = importlib.util.spec_from_file_location(
            scheduler[0].split("/")[-1].split(".")[0], scheduler[0]
        )
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
        schedule_handler_class = getattr(module, scheduler[1])
        if not astroplan.Scheduler in schedule_handler_class.__bases__:
            logger.error(
                f"Scheduler {scheduler[0]} does not inherit from astroplan.Scheduler."
            )
            return
        schedule_handler = schedule_handler_class(
            constraints=global_constraints,
            observer=observatory,
            transitioner=transitioner,
            gap_time=gap_time * u.second,
            time_resolution=time_resolution * u.second,
        )

    logger.info("Scheduling ObservingBlocks")
    for i in range(len(blocks)):
        logger.debug("Block group %i of i" % (i + 1, len(blocks)))
        schedule_handler(blocks[i], schedule)

    # Generate schedule table
    schedule_table = schedule.to_table(show_transitions=False, show_unused=False)

    # Update ephem for non-sidereal targets
    for row in schedule_table:
        if (
            row["configuration"]["pm_ra_cosdec"].value != 0
            or row["configuration"]["pm_dec"].value != 0
        ):
            logger.info("Updating ephemeris for %s at scheduled time" % row["target"])
            ephemerides = mpc.MPC.get_ephemeris(
                target=row["target"],
                location=observatory.location,
                start=row["start time (UTC)"],
                number=1,
                proper_motion="sky",
            )
            row["ra"] = ephemerides["RA"][0]
            row["dec"] = ephemerides["DEC"][0]
            row["configuration"]["pm_ra_cosdec"] = (
                ephemerides["dRA cos(Dec)"][0] * u.arcsec / u.hour
            )
            row["configuration"]["pm_dec"] = ephemerides["dDec"][0] * u.arcsec / u.hour

    # Re-assign filenames
    name_dict = {}
    for i in range(len(schedule_table)):
        name = schedule_table[i]["configuration"]["obscode"]

        if name in name_dict:
            name_dict[name] += 1
        else:
            name_dict[name] = 0

        name += (
            ("%3.3g" % date.strftime("%j")) + "_" + ("%4.4g" % name_dict[name]) + ".fts"
        )
        schedule_table[i]["configuration"]["filename"] = name

    # Write the telrun.ecsv file
    logger.info("Writing schedule to file")
    if filename is None or telrun:
        first_time = astrotime.Time(
            schedule_table[0]["start time (UTC)"], format="iso", scale="utc"
        ).strftime("%m-%d-%Y")
        filename = "telrun_" + first_time + ".ecsv"

    if telrun:
        try:
            path = os.environ.get("TELRUN_EXECUTE")
            logger.info(
                "-t/--telrun flag set, writing schedule to %s from $TELRUN_EXECUTE environment variable"
                % path
            )
        except:
            path = os.getcwd() + "/schedules/"
            logger.info("-t/--telrun flag set, writing schedule to %s" % path)
        if not os.path.isdir(path):
            logger.exception(f"Path {path} does not exist.")
            return
    else:
        path = os.getcwd() + "/"
        logger.info("-t/--telrun flag not set, writing schedule to %s" % path)
    schedule_table.write(path + filename, format="ascii.ecsv", overwrite=True)

    # Plot the schedule
    ax = None
    match plot:
        case 1:  # Gantt chart
            logger.info("Plotting schedule as a Gantt chart")
            ax = plot_schedule_gantt(schedule, observatory, name=instrument_name)
            return schedule_table, ax
        case 2:  # Plot by target w/ altitude
            logger.info("Plotting schedule by target with airmass")
            ax = astroplan.plots.plot_schedule_airmass(schedule)
            plt.legend()
            return schedule_table, ax
        case 3:  # Sky chart
            logger.info("Plotting schedule on a sky chart")
            objects = []
            times = []
            for i in range(len(schedule_table)):
                if schedule_table[i]["ra"] not in [obj.ra.dms for obj in objects]:
                    objects.append(
                        coord.SkyCoord(
                            ra=schedule_table[i]["ra"], dec=schedule_table[i]["dec"]
                        )
                    )
                    times.append(
                        [
                            (
                                astrotime.Time(
                                    schedule_table[i]["start time (UTC)"],
                                    format="iso",
                                    scale="utc",
                                )
                                + astrotime.Time(
                                    schedule_table[i]["end time (UTC)"],
                                    format="iso",
                                    scale="utc",
                                )
                            )
                            / 2
                        ]
                    )
                elif schedule_table[i]["dec"] != [obj.ra.dms for obj in objects].index(
                    schedule_table[i]["ra"]
                ):
                    objects.append(
                        coord.SkyCoord(
                            ra=schedule_table[i]["ra"], dec=schedule_table[i]["dec"]
                        )
                    )
                    times.append(
                        [
                            (
                                astrotime.Time(
                                    schedule_table[i]["start time (UTC)"],
                                    format="iso",
                                    scale="utc",
                                )
                                + astrotime.Time(
                                    schedule_table[i]["end time (UTC)"],
                                    format="iso",
                                    scale="utc",
                                )
                            )
                            / 2
                        ]
                    )
                else:
                    times[
                        [obj.dec.dms for obj in objects].index(schedule_table[i]["dec"])
                    ].append(
                        (
                            astrotime.Time(
                                schedule_table[i]["start time (UTC)"],
                                format="iso",
                                scale="utc",
                            )
                            + astrotime.Time(
                                schedule_table[i]["end time (UTC)"],
                                format="iso",
                                scale="utc",
                            )
                        )
                        / 2
                    )

            for i in range(len(objects)):
                ax = astroplan.plot_sky(
                    astroplan.FixedTarget(objects[i]),
                    observatory,
                    times[i],
                    style_kwargs={
                        "color": ccm.batlow(i / (len(objects) - 1)),
                        "label": objects[i].to_string("hmsdms"),
                    },
                )
            plt.legend()
            return schedule_table, ax
        case _:
            logger.info("No plot requested")
            pass

    return schedule_table


@click.command()
@click.argument("schedule_table", type=click.Path(exists=True))
@click.argument("observatory", type=click.Path(exists=True))
@click.option("-n", "--name", type=str, default="", help="Name of observatory.")
@click.version_option(version="0.1.0")
@click.help_option("-h", "--help")
def plot_schedule_gantt_cli(schedule_table, observatory, name):
    if type(schedule_table) is not table.Table:
        schedule_table = table.Table.read(schedule_table, format="ascii.ecsv")

    if type(observatory) is not astroplan.Observer:
        if type(observatory) is str:
            obs_cfg = configparser.ConfigParser()
            obs_cfg.read(observatory)
            location = coord.EarthLocation(
                lon=obs_cfg.getfloat("location", "longitude") * u.deg,
                lat=obs_cfg.getfloat("location", "latitude") * u.deg,
                height=obs_cfg.getfloat("location", "elevation") * u.m,
            )
            name = obs_cfg.get("site", "instrument_name")
            observatory = astroplan.Observer(location=location)
        elif type(observatory) is Observatory:
            location = observatory.observatory_location
            name = observatory.instrument_name
            observatory = astroplan.Observer(location=observatory.observatory_location)
        else:
            raise TypeError(
                "Observatory must be, a string, Observatory object, or astroplan.Observer object."
            )
    else:
        location = observatory.location
        name = observatory.name

    obscodes = np.unique(
        [block["configuration"]["obscode"] for block in schedule_table]
    )

    date = astrotime.Time(
        schedule_table[0]["start time (UTC)"], format="iso", scale="utc"
    ).datetime
    t0 = (
        astrotime.Time(
            datetime.datetime(date.year, date.month, date.day, 12, 0, 0),
            format="datetime",
            scale="utc",
        )
        - (
            (
                astrotime.Time(date, format="datetime", scale="utc")
                - astrotime.Time.now()
            ).day
            % 1
        )
        * u.day
    )

    fig, ax = plt.subplots(1, 1, figsize=(12, len(obscodes) / 2))

    for i in range(len(obscodes)):
        plot_blocks = [
            block
            for block in schedule_table
            if block["configuration"]["obscode"] == obscodes[i]
        ]

        for block in plot_blocks:
            start_time = astrotime.Time(
                block["start time (UTC)"], format="iso", scale="utc"
            )
            end_time = astrotime.Time(
                block["end time (UTC)"], format="iso", scale="utc"
            )
            length_minutes = int((start_time - end_time).sec / 60 + 1)
            times = (
                start_time
                + np.linspace(0, length_minutes, length_minutes, endpoint=True)
                * u.minute
            )

            obj = coord.SkyCoord(ra=block["ra"], dec=block["dec"]).transform_to(
                coord.AltAz(obstime=times, location=location)
            )
            airmass = utils.airmass(obj.alt * u.deg)

            ax.scatter(
                times.datetime,
                i * np.ones(len(times)),
                lw=0,
                marker="s",
                c=airmass,
                cmap=ccm.batlow,
                vmin=1,
                vmax=3,
            )

            scatter = ax.scatter(
                times.datetime,
                len(obscodes) * np.ones(len(times)),
                lw=0,
                marker="s",
                c=airmass,
                cmap=ccm.batlow,
                vmin=1,
                vmax=3,
            )

    obscodes.append("All")

    twilight_times = [
        t0,
        astrotime.Time(observatory.sun_set_time(t0, which="next"), scale="utc"),
        astrotime.Time(
            observatory.twilight_evening_civil(t0, which="next"), scale="utc"
        ),
        astrotime.Time(
            observatory.twilight_evening_nautical(t0, which="next"), scale="utc"
        ),
        astrotime.Time(
            observatory.twilight_evening_astronomical(t0, which="next"), scale="utc"
        ),
        astrotime.Time(
            observatory.twilight_morning_astronomical(t0, which="next"), scale="utc"
        ),
        astrotime.Time(
            observatory.twilight_morning_nautical(t0, which="next"), scale="utc"
        ),
        astrotime.Time(
            observatory.twilight_morning_civil(t0, which="next"), scale="utc"
        ),
        astrotime.Time(observatory.sun_rise_time(t0, which="next"), scale="utc"),
        t1,
    ]
    opacities = [0.8, 0.6, 0.4, 0.2, 0, 0.2, 0.4, 0.6, 0.8]
    for i in range(len(twilight_times) - 1):
        ax.axvspan(
            twilight_times[i].datetime,
            twilight_times[i + 1].datetime,
            color="grey",
            alpha=opacities[i],
        )

    ax.set_xlabel(
        "Time beginning %s [UTC]"
        % (twilight_times[1] - 0.5 * u.hour).strftime("%m-%d-%Y")
    )
    ax.set_xlim(
        [
            (twilight_times[1] - 0.5 * u.hour).datetime,
            (twilight_times[-2].datetime + 0.5 * u.hour).datetime,
        ]
    )
    ax.xaxis.set_major_formatter(mdates.DateFormatter("%H:%M"))
    ax.xaxis.set_major_locator(mdates.HourLocator(interval=1))
    ax.xaxis.set_minor_formatter(ticks.NullFormatter())
    ax.xaxis.set_minor_locator(mdates.MinuteLocator(interval=15))
    ax.xtick_params(rotation=45)

    ax.set_ylabel("Observer Code")
    ax.set_ylim([len(obscodes) + 1.5, 0.5])
    ax.yaxis.set_major_formatter(ticker.FixedFormatter(obscodes))
    ax.yaxis.set_major_locator(ticker.MultipleLocator(1))
    ax.yaxis.set_minor_formatter(ticks.NullFormatter())
    ax.yaxis.set_minor_locator(ticker.NullLocator())

    cbar = plt.colorbar(scatter, ax=ax)
    cbar.set_yticks([1, 1.5, 2, 2.5, 3])
    cbar.set_label("Airmass", rotation=270, labelpad=20)

    ax.set_title(
        "Observing Schedule for %s on %s"
        % (name, twilight_times[1] - 0.5 * u.hour).strftime("%m-%d-%Y")
    )
    ax.grid()

    return ax


def parse_sch_file(filename, location=None, t0=None):
    with open(filename, "r") as f:
        raw_lines = f.readlines()

    # Remove equal signs, quotes, and blank lines
    lines = []
    for line in raw_lines:
        logger.debug(f"Parsing line: {line}")
        line = line.replace("=", " ")
        line = line.replace("`", "'")
        line = line.replace('"', "'")
        line = line.replace("‘", "'")
        line = line.replace("’", "'")

    lines = [line.split("#")[0] for line in lines]  # Remove comments
    lines = [line.split("!")[0] for line in lines]  # Remove comments
    lines = [line.replace("\n", "") for line in lines]  # Remove line breaks
    " ".join(
        mystring.split()
    )  # Remove multiple spaces, trailing and leading whitespace
    lines = [line for line in lines if line != ""]  # Remove empty lines
    lines = [line.lower() for line in lines]  # Lower case

    # Look for title, observer keywords
    title = ""
    for line in lines:
        title = _get_keyvalue(line, "tit", start_line=True)
        if title is not None:
            lines.remove(line)
            break

    observer = ""
    for line in lines:
        val += _get_keyvalue(line, "obs", start_line=True)
        if val is not None:
            lines.remove(line)
            observer += val + ","

    code = filename[:3]

    prior_filters = None
    prior_exposures = None

    # Parse each line and place into ObservingBlock
    blocks = []

    for line in lines:
        # Required keywords
        target_name = _get_keyvalue(line, ("tar", "sou", "obj", "nam"))
        ra = _get_keyvalue(line, "ra")
        dec = _get_keyvalue(line, "dec")
        nonsidereal = _get_flag(line, "non")

        if None not in (ra, dec):
            obj = coord.SkyCoord(ra, dec, unit=(u.hourangle, u.deg))
            if target_name is None:
                target_name = obj.to_string("hmsdms")
            obj = astroplan.FixedTarget(name=nm, coord=obj)
            pm_ra_cosdec = telrun.observing_block_config["pm_ra_cosdec"]
            pm_dec = telrun.observing_block_config["pm_dec"]
        elif target_name != None and not nonsidereal:
            obj = coord.SkyCoord.from_name(target_name)
            obj = astroplan.FixedTarget(name=target, coord=obj)
            pm_ra_cosdec = telrun.observing_block_config["pm_ra_cosdec"]
            pm_dec = telrun.observing_block_config["pm_dec"]
        elif target_name != None and nonsidereal:
            ephemerides = mpc.MPC.get_ephemeris(
                target=target_name,
                location=location,
                start=t0 + 0.5 * u.day,
                number=1,
                proper_motion="sky",
            )
            ra = ephemerides["RA"][0]
            dec = ephemerides["DEC"][0]
            pm_ra_cosdec = ephemerides["dRA cos(Dec)"][0] * u.arcsec / u.hour
            pm_dec = ephemerides["dDec"][0] * u.arcsec / u.hour
            obj = coord.SkyCoord(ra, dec, pm_ra_cosdec=pm_ra_cosdec, pm_dec=pm_dec)
            obj = astroplan.FixedTarget(name=target_name, coord=obj)
        else:
            pass  # TODO: allow for sources with no coordinates, i.e. scheduling darks/flats

        # Get non-iterating, non-inheriting keywords
        name = _get_keyvalue(line, "file")
        if name is None:
            name = telrun.observing_block_config["filename"]

        priority = _get_keyvalue(line, "pri")
        priority = int(priority) if priority is not None else 1

        repositioning = _get_keyvalue(line, "repo")
        if repositioning is not None:
            repositioning = (
                int(repositioning.split(",")[0]),
                int(repositioning.split(",")[1]),
            )
        else:
            repositioning = telrun.observing_block_config["repositioning"]

        shutter_state = _get_keyvalue(line, "shu", boolean=True)
        if shutter_state is None:
            shutter_state = telrun.observing_block_config["shutter_state"]

        readout = _get_keyvalue(line, "rea")
        if readout is not None:
            readout = int(readout)
        else:
            readout = telrun.observing_block_config["readout"]

        binning = _get_keyvalue(line, "bin")
        if binning is not None:
            binning = (int(binning.split("x")[0]), int(binning.split("x")[1]))
        else:
            binning = telrun.observing_block_config["binning"]

        frame_position = _get_keyvalue(line, "frame_p")
        if frame_position is not None:
            frame_position = (
                int(frame_position.split(",")[0]),
                int(frame_position.split(",")[1]),
            )
        else:
            frame_position = telrun.observing_block_config["frame_position"]

        frame_size = _get_keyvalue(line, "frame_s")
        if frame_size is not None:
            frame_size = (int(frame_size.split(",")[0]), int(frame_size.split(",")[1]))
        else:
            frame_size = telrun.observing_block_config["frame_size"]

        comment = _get_keyvalue(line, "com")
        if comment is None:
            comment = telrun.observing_block_config["comment"]

        # Get timing keywords
        utstart = _get_keyvalue(line, "uts", "sta")
        if utstart is not None:
            utstart = astrotime.Time(utstart, format="isot", scale="utc")

        cadence = _get_keyvalue(line, "cad")
        if cadence is not None:
            cadence = astrotime.TimeDelta(
                datetime.time(*[int(c) for c in cadence.split(":")]), format="datetime"
            )

            if utstart is None:
                raise ValueError("Must specify utstart if cadence is specified")

        schederr = _get_keyvalue(line, "sch")
        if schederr is not None:
            schederr = astrotime.TimeDelta(
                datetime.time(*[int(s) for s in schederr.split(":")]), format="datetime"
            )

        if utstart is None and schederr is not None:
            raise ValueError("Must specify utstart if schederr is specified")
        elif utstart is not None and schederr is None:
            schederr = 60 * u.second

        # Get exposure behavior keywords
        n_exp = _get_keyvalue(line, ("n_e", "nex", "repe"))
        if n_exp is not None:
            n_exp = int(n_exp)
        else:
            n_exp = telrun.observing_block_config["n_exp"]

        do_not_interrupt = _get_flag(line, ("don", "do_", "do-"))
        if do_not_interrupt is None:
            do_not_interrupt = telrun.observing_block_config["do_not_interrupt"]

        # Get iterating, inheriting keywords
        filters = _get_keyvalue(line, "filt")
        if filters is not None:
            filters = filters.split(",")
            prior_filters = filters
        elif prior_filters is not None:
            filters = prior_filters
        else:
            filters = telrun.observing_block_config["filters"]
            prior_filters = None

        exposures = _get_keyvalue(line, "exp")
        if exposures is not None:
            exposures = [float(e) for e in exposures.split(",")]
            prior_exposures = exposures
        elif prior_exposures is not None:
            exposures = prior_exposures
        else:
            exposures = telrun.observing_block_config["exposures"]
            prior_exposures = None

        # Sanity Check 1: matching number of filters and exposures
        if len(filters) != len(exposures) and len(filters) != 0:
            raise ValueError("Number of filters must match number of exposures")

        # Sanity Check 2: do_not_interrupt and cadence don't both appear:
        if do_not_interrupt is not None and cadence is not None:
            raise ValueError(
                "Cannot specify do_not_interrupt and cadence simultaneously"
            )

        # Sanity Check 3: if cadence is specified, verify it exceeds exposure time
        # times number of exposures times number of filters
        if cadence is not None:
            if cadence < np.sum(exposures) * n_exp * len(filters):
                raise ValueError("Cadence must be greater than total exposure time")

        for i in range(len(filters)):
            filt = filters[i]
            exp = exposures[i]
            constraints = None

            if do_not_interrupt:
                loop_max = 1
                temp_dur = exp * n_exp * u.second
                temp_n_exp = n_exp
            else:
                loop_max = n_exp
                temp_dur = exp * u.second
                temp_n_exp = 1

            if utstart is not None:
                if cadence is not None:
                    constraint_cadence = cadence
                else:
                    constraint_cadence = temp_dur

                constraints = [
                    [
                        astroplan.constraints.TimeConstraint(
                            utstart + (i + j * len(i)) * constraint_cadence - schederr,
                            utstart + (i + j * len(i)) * constraint_cadence + schederr,
                        )
                    ]
                    for j in range(loop_max)
                ]

            for j in range(loop_max):
                blocks.append(
                    astroplan.ObservingBlock(
                        target=obj,
                        duration=temp_dur,
                        priority=priority,
                        name=target_name,
                        configuration={
                            "observer": observer,
                            "code": code,
                            "title": title,
                            "filename": name,
                            "filter": filt,
                            "exposure": exp,
                            "n_exp": temp_n_exp,
                            "do_not_interrupt": do_not_interrupt,
                            "repositioning": repositioning,
                            "shutter_state": shutter_state,
                            "readout": readout,
                            "binning": binning,
                            "frame_position": frame_position,
                            "frame_size": frame_size,
                            "pm_ra_cosdec": pm_ra_cosdec,
                            "pm_dec": pm_dec,
                            "comment": comment,
                            "status": "N",
                            "message": "unprocessed",
                        },
                        constraints=constraints[j],
                    )
                )
                logger.debug(
                    f"""Created ObservingBlock: {blocks[-1].target},
                                {blocks[-1].duration}, {blocks[-1].priority},
                                {blocks[-1].name}, {blocks[-1].constraints},
                                {blocks[-1].configuration}"""
                )

    return blocks


def _get_keyvalue(line, keyword, start_line=False, boolean=False):
    if start_line:
        if line.lower().startswith(keyword.lower()):
            if not boolean:
                return shlex.split(line)[1]
            else:
                if shlex.split(line)[1].lower().startswith(("t", "y", "1")):
                    return True
                elif shlex.split(line)[1].lower().startswith(("f", "n", "0")):
                    return False
                else:
                    return None
        else:
            return None

    for i in range(len(line)):
        if line[i:].lower().startswith(keyword.lower()):
            return shlex.split(line[i:])[1]
        else:
            return None


def _get_flag(line, keyword):
    for i in range(len(line)):
        if line[i:].lower().startswith(keyword.lower()):
            return True

    return False


schedtel = schedtel_cli.callback
plot_schedule_gantt = plot_schedule_gantt_cli.callback



import asyncio

import pytest

from pyscope.utils import _force_async, _force_sync


def test__forceaync():
    @_force_async
    def t():
        import time

        time.sleep(1)
        return True

    @_force_sync
    async def main():
        import asyncio

        # if it were to execute sequentially, it would take 10 seconds, in this case we expect to see only 1 second
        futures = list(map(lambda x: t(), range(10)))
        return await asyncio.gather(*futures)

    assert main() == [True] * 10


def test__forcesync():
    @_force_sync
    async def fn():
        return 1

    assert fn() == 1



import asyncio
import functools


def _force_async(fn):
    """
    turns a sync function to async function using threads
    """
    import asyncio
    from concurrent.futures import ThreadPoolExecutor

    pool = ThreadPoolExecutor()

    @functools.wraps(fn)
    def wrapper(*args, **kwargs):
        future = pool.submit(fn, *args, **kwargs)
        return asyncio.wrap_future(future)  # make it awaitable

    return wrapper


def _force_sync(fn):
    """
    turn an async function to sync function
    """
    import asyncio

    @functools.wraps(fn)
    def wrapper(*args, **kwargs):
        res = fn(*args, **kwargs)
        if asyncio.iscoroutine(res):
            return asyncio.get_event_loop().run_until_complete(res)
        return res

    return wrapper



import logging

from .ascom_device import ASCOMDevice
from .focuser import Focuser

logger = logging.getLogger(__name__)


class ASCOMFocuser(ASCOMDevice, Focuser):
    def __init__(self, identifier, alpaca=False, device_number=0, protocol="http"):
        super().__init__(
            identifier,
            alpaca=alpaca,
            device_type=self.__class__.__name__.replace("ASCOM", ""),
            device_number=device_number,
            protocol=protocol,
        )

    def Halt(self):
        logger.debug(f"ASCOMFocuser.Halt() called")
        self._device.Halt()

    def Move(self, Position):
        logger.debug(f"ASCOMFocuser.Move({Position}) called")
        self._device.Move(Position)

    @property
    def Absolute(self):
        logger.debug(f"ASCOMFocuser.Absolute property called")
        return self._device.Absolute

    @property
    def IsMoving(self):
        logger.debug(f"ASCOMFocuser.IsMoving property called")
        return self._device.IsMoving

    @property
    def MaxIncrement(self):
        logger.debug(f"ASCOMFocuser.MaxIncrement property called")
        return self._device.MaxIncrement

    @property
    def MaxStep(self):
        logger.debug(f"ASCOMFocuser.MaxStep property called")
        return self._device.MaxStep

    @property
    def Position(self):
        logger.debug(f"ASCOMFocuser.Position property called")
        return self._device.Position

    @property
    def StepSize(self):
        logger.debug(f"ASCOMFocuser.StepSize property called")
        return self._device.StepSize

    @property
    def TempComp(self):
        logger.debug(f"ASCOMFocuser.TempComp property called")
        return self._device.TempComp

    @TempComp.setter
    def TempComp(self, value):
        logger.debug(f"ASCOMFocuser.TempComp property set to {value}")
        self._device.TempComp = value

    @property
    def TempCompAvailable(self):
        logger.debug(f"ASCOMFocuser.TempCompAvailable property called")
        return self._device.TempCompAvailable

    @property
    def Temperature(self):
        logger.debug(f"ASCOMFocuser.Temperature property called")
        return self._device.Temperature
